using System;
using System.Text;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.crypto.modes;
using org.bouncycastle.util.encoders;
using org.bouncycastle.util.test;


namespace test.crypto
{
	/**
	* CTS tester
	*/
	public class CTSTest: Test
	{
		static byte[]   in1 = Hex.decode("4e6f7720697320746865207420");
		static byte[]   in2 = Hex.decode("000102030405060708090a0b0c0d0e0fff0102030405060708090a0b0c0d0e0f0aaa");
		static byte[]   out1 = Hex.decode("9952f131588465033fa40e8a98");
		static byte[]   out2 = Hex.decode("358f84d01eb42988dc34efb994");
		static byte[]   out3 = Hex.decode("170171cfad3f04530c509b0c1f0be0aefbd45a8e3755a873bff5ea198504b71683c6");

		private class CTSTester: Test
		{
			private int                 id;
			private BlockCipher         cipher;
			private CipherParameters    parameters;
			private byte[]              input;
			private byte[]              output;
	    
			internal CTSTester(
				int                 id,
				BlockCipher         cipher,
				CipherParameters    parameters,
				byte[]              input,
				byte[]              output)
			{
				this.id = id;
				this.cipher = cipher;
				this.parameters = parameters;
				this.input = input;
				this.output = output;
			}
	    
			public String getName()
			{
				return "CTSTest " + id;
			}
	    
			public TestResult perform()
			{
				byte[]                  outBytes = new byte[input.Length];
				BufferedBlockCipher     engine = new CTSBlockCipher(cipher);
	    
				engine.init(true, parameters);

   
				int len = engine.processBytes(input, 0, input.Length, outBytes, 0);
	    
				try
				{
					engine.doFinal(outBytes, len);
				}
				catch (Exception e)
				{
					return new SimpleTestResult(false, getName() + ": encryption exception - " + e.ToString());
				}
	    
				for (int i = 0; i != output.Length; i++)
				{
					if (outBytes[i] != output[i])
					{
						return new SimpleTestResult(false, getName() + ": failed encryption expected " + ByteArray2String(Hex.encode(output)) + " got " + ByteArray2String(Hex.encode(outBytes)));
					}
				}
	    
				engine.init(false, parameters);
	    
				len = engine.processBytes(output, 0, output.Length, outBytes, 0);
	    
				try
				{
					engine.doFinal(outBytes, len);
				}
				catch (Exception e)
				{
					return new SimpleTestResult(false, getName() + ": decryption exception - " + e.ToString());
				}
	    
				for (int i = 0; i != input.Length; i++)
				{
					if (outBytes[i] != input[i])
					{
						return new SimpleTestResult(false, getName() + ": failed encryption expected " + ByteArray2String(Hex.encode(input)) + " got " + ByteArray2String(Hex.encode(outBytes)));
					}
				}
	    
				return new SimpleTestResult(true, getName() + ": Okay");
			}
			private String ByteArray2String(byte[] bs)
			{
				StringBuilder s = new StringBuilder();
				for (int i = 0; i < bs.Length; i++) 
				{
					s.Append(Convert.ToChar(bs[i]));
				}
				return s.ToString();
			}
		}
	    
		internal CTSTest()
		{
		}

		public String getName()
		{
			return "CTS";
		}

		public TestResult perform()
		{
			byte[]  key1 = { (byte)0x01, (byte)0x23, (byte)0x45, (byte)0x67, (byte)0x89, (byte)0xAB, (byte)0xCD, (byte)0xEF };
			byte[]  key2 = { (byte)0x01, (byte)0x23, (byte)0x45, (byte)0x67, (byte)0x89, (byte)0xAB, (byte)0xCD, (byte)0xEF, (byte)0xee, (byte)0xff  };
			byte[]  iv = { 1, 2, 3, 4, 5, 6, 7, 8 };

			Test        test = new CTSTester(1, new DESEngine(), new KeyParameter(key1), in1, out1);
			TestResult  result = test.perform();

			if (!result.isSuccessful())
			{
				return result;
			}

			test = new CTSTester(2, new CBCBlockCipher(new DESEngine()), new ParametersWithIV(new KeyParameter(key1), iv), in1, out2);
			result = test.perform();

			if (!result.isSuccessful())
			{
				return result;
			}

			test = new CTSTester(3, new CBCBlockCipher(new SkipjackEngine()), new ParametersWithIV(new KeyParameter(key2), iv), in2, out3);
			result = test.perform();

			if (!result.isSuccessful())
			{
				return result;
			}

			return new SimpleTestResult(true, getName() + ": Okay");
		}

		[STAThread]
		static void Main(string[] args)
		{
			CTSTest         test = new CTSTest();
			TestResult      result = test.perform();

			Console.WriteLine(result.ToString());
		}

	}

}
