using System;
using System.IO;
using System.Collections;
//using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	public abstract class ASN1Sequence : DERObject
	{
		private ArrayList seq = new ArrayList();

		/**
		* return an ASN1Sequence from the given object.
		*
		* @param obj the object we want converted.
		* @exception ArgumentException if the object cannot be converted.
		*/
		public static ASN1Sequence getInstance(
			Object  obj)
		{
			if ((obj == null) || (typeof(ASN1Sequence).IsInstanceOfType(obj)))
			{
				return (ASN1Sequence)obj;
			}
			throw new ArgumentException("unknown object in getInstance");
		}

		/**
		* Return an ASN1 sequence from a tagged object. There is a special
		* case here, if an object appears to have been explicitly tagged on 
		* reading but we were expecting it to be implictly tagged in the 
		* normal course of events it indicates that we lost the surrounding
		* sequence - so we need to add it back (this will happen if the tagged
		* object is a sequence that contains other sequences). If you are
		* dealing with implicitly tagged sequences you really <b>should</b>
		* be using this method.
		*
		* @param obj the tagged object.
		* @param flagExplicit true if the object is meant to be explicitly tagged,
		*          false otherwise.
		* @exception ArgumentException if the tagged object cannot
		*          be converted.
		*/
		public static ASN1Sequence getInstance(
			ASN1TaggedObject    obj,
			bool             flagExplicit)
		{
			if (flagExplicit)
			{
				if (!obj.isExplicit())
				{
					throw new ArgumentException("object implicit - explicit expected.");
				}

				return (ASN1Sequence)obj.getObject();
			}
			else
			{
				//
				// constructed object which appears to be explicitly tagged
				// when it should be implicit means we have to add the
				// surrounding sequence.
				//
				if (obj.isExplicit())
				{
					if (typeof(BERTaggedObject).IsInstanceOfType(obj))
					{
						return new BERSequence(obj.getObject());
					}
					else
					{
						return new DERSequence(obj.getObject());
					}
				}
				else
				{
					if (typeof(ASN1Sequence).IsInstanceOfType(obj.getObject()))
					{
						return (ASN1Sequence)obj.getObject();
					}
				}
			}

			throw new ArgumentException(
					"unknown object in getInstanceFromTagged");
		}

		public virtual IEnumerator getObjects()
		{
			return seq.GetEnumerator();
		}

		/**
		* return the object at the sequence postion indicated by index.
		*
		* @param the sequence number (starting at zero) of the object
		* @return the object at the sequence postion indicated by index.
		*/
		public DEREncodable getObjectAt(
			int index)
		{
			return (DEREncodable)seq[index];
		}

		/**
		* return the number of objects in this sequence.
		*
		* @return the number of objects in this sequence.
		*/
		public virtual int size()
		{
			return seq.Count;
		}

		public override int GetHashCode()
		{
			IEnumerator             e = this.getObjects();
			int                     hashCode = 0;

			while (e.MoveNext())
			{
				hashCode ^= e.Current.GetHashCode();
			}

			return hashCode;
		}

		public override bool Equals(
			Object  o)
		{
			if ((o == null) || !(typeof(ASN1Sequence).IsInstanceOfType(o)))
			{
				return false;
			}

			ASN1Sequence   other = (ASN1Sequence)o;

			if (this.size() != other.size())
			{
				return false;
			}

			IEnumerator s1 = this.getObjects();
			IEnumerator s2 = other.getObjects();

			while (s1.MoveNext()&& s2.MoveNext())
			{
				if (!s1.Current.Equals(s2.Current))
				{
					return false;
				}
			}

			return true;
		}

		public virtual void addObject(
			DEREncodable obj)
		{
			seq.Add(obj);
		}
	}

}