using System;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.digests;
using org.bouncycastle.crypto.parameters;

namespace org.bouncycastle.crypto.signers
{
	/**
	* ISO9796-2 - mechanism using a hash function.
	*/
	public class ISO9796d2Signer
	{
		static readonly public int   TRAILER_IMPLICIT    = 0xBC;
		static readonly public int   TRAILER_RIPEMD160   = 0x31CC;
		static readonly public int   TRAILER_RIPEMD128   = 0x32CC;
		static readonly public int   TRAILER_SHA1        = 0x33CC;

		private Digest                      digest;
		private AsymmetricBlockCipher       cipher;

		private int         trailer;
		private int         keyBits;
		private byte[]      block;
		private byte[]      mBuf;
		private int         messageLength;

		public ISO9796d2Signer(
			AsymmetricBlockCipher   cipher,
			Digest                  digest,
			bool                 flagImplicit)
		{
			this.cipher = cipher;
			this.digest = digest;

			if (flagImplicit)
			{
				trailer = TRAILER_IMPLICIT;
			}
			else
			{
				
				if (typeof(SHA1Digest).IsInstanceOfType(digest))
				{
					trailer = TRAILER_SHA1;
				}
				else if (typeof(RIPEMD160Digest).IsInstanceOfType(digest))
				{
					trailer = TRAILER_RIPEMD160;
				}
				else if (typeof(RIPEMD128Digest).IsInstanceOfType(digest))
				{
					trailer = TRAILER_RIPEMD128;
				}
				else
				{
					throw new ArgumentException("no valid trailer for digest");
				}
			}
		}

		public ISO9796d2Signer(
			AsymmetricBlockCipher   cipher,
			Digest                  digest): this(cipher, digest, false){}

		public void init(
			bool                 forSigning,
			CipherParameters        param)
		{
			RSAKeyParameters  kParam = (RSAKeyParameters)param;

			cipher.init(forSigning, kParam);

			keyBits = kParam.getModulus().bitLength();

			block = new byte[(keyBits + 7) / 8];
			mBuf = new byte[block.Length - 1];

			reset();
		}

		/**
		* clear possible sensitive data
		*/
		private void clearBlock(
			byte[]  block)
		{
			for (int i = 0; i != block.Length; i++)
			{
				block[i] = 0;
			}
		}

		/**
		* update the internal digest with the byte b
		*/
		public void update(
			byte    b)
		{
			digest.update(b);

			if (messageLength < mBuf.Length)
			{
				mBuf[messageLength] = b;
			}

			messageLength++;
		}

		/**
		* update the internal digest with the byte array in
		*/
		public void update(
			byte[]  inBytes,
			int     off,
			int     len)
		{
			digest.update(inBytes, off, len);

			if (messageLength < mBuf.Length)
			{
				for (int i = 0; i < len && (i + messageLength) < mBuf.Length; i++)
				{
					mBuf[messageLength + i] = inBytes[off + i];
				}
			}

			messageLength += len;
		}

		/**
		* reset the internal state
		*/
		public void reset()
		{
			digest.reset();
			messageLength = 0;
		}

		/**
		* generate a signature for the loaded message using the key we were
		* initialised with.
		*/
		public byte[] generateSignature()
			//throws CryptoException
		{
			int     digSize = digest.getDigestSize();

			int t = 0;
			int delta = 0;

			if (trailer == TRAILER_IMPLICIT)
			{
				t = 8;
				delta = block.Length - digSize - 1;
				digest.doFinal(block, delta);
				block[block.Length - 1] = (byte)TRAILER_IMPLICIT;
			}
			else
			{
				t = 16;
				delta = block.Length - digSize - 2;
				digest.doFinal(block, delta);
				block[block.Length - 2] = (byte)((uint) trailer >> 8);
				block[block.Length - 1] = (byte)trailer;
			}

			byte    header = 0;
			int     x = (digSize + messageLength) * 8 + t + 4 - keyBits;

			if (x > 0)
			{
				int mR = messageLength - ((x + 7) / 8);
				header = 0x60;

				delta -= mR;
	            
				Array.Copy(mBuf, 0, block, delta, mR);
			}
			else
			{
				header = 0x40;
				delta -= messageLength;
	            
				Array.Copy(mBuf, 0, block, delta, messageLength);
			}
	        
			if ((delta - 1) > 0)
			{
				for (int i = delta - 1; i != 0; i--)
				{
					block[i] = (byte)0xbb;
				}
				block[delta - 1] ^= (byte)0x01;
				block[0] = (byte)0x0b;
				block[0] |= header;
			}
			else
			{
				block[0] = (byte)0x0a;
				block[0] |= header;
			}

			byte[]  b = cipher.processBlock(block, 0, block.Length);

			clearBlock(mBuf);
			clearBlock(block);

			return b;
		}

		/**
		* return true if the signature represents a ISO9796-2 signature
		* for the passed in message.
		*/
		public bool verifySignature(
			byte[]      signature)
		{
			byte[]      block = null;

			try
			{
				block = cipher.processBlock(signature, 0, signature.Length);
			}
			catch (Exception e)
			{
				e.ToString();
				return false;
			}

			if (((block[0] & 0xC0) ^ 0x40) != 0)
			{
				clearBlock(mBuf);
				clearBlock(block);

				return false;
			}

			if (((block[block.Length - 1] & 0xF) ^ 0xC) != 0)
			{
				clearBlock(mBuf);
				clearBlock(block);

				return false;
			}

			int     delta = 0;

			if (((block[block.Length - 1] & 0xFF) ^ 0xBC) == 0)
			{
				delta = 1;
			}
			else
			{
				int sigTrail = ((block[block.Length - 2] & 0xFF) << 8) | (block[block.Length - 1] & 0xFF);

				if (sigTrail == TRAILER_RIPEMD160)
				{
					if (!(typeof(RIPEMD160Digest).IsInstanceOfType(digest)))
					{
						throw new Exception("signer should be initialised with RIPEMD160");
					}
				}
				else if (sigTrail == TRAILER_SHA1)
				{
					if (!(typeof(SHA1Digest).IsInstanceOfType(digest)))
					{
						throw new Exception("signer should be initialised with SHA1");
					}
				}
				else if (sigTrail == TRAILER_RIPEMD128)
				{
					if (!(typeof(RIPEMD128Digest).IsInstanceOfType(digest)))
					{
						throw new Exception("signer should be initialised with RIPEMD128");
					}
				}
				else
				{
					throw new ArgumentException("unrecognised hash in signature");
				}

				delta = 2;
			}

			//
			// find out how much padding we've got
			//
			int mStart = 0;

			for (mStart = 0; mStart != block.Length; mStart++)
			{
				if (((block[mStart] & 0x0f) ^ 0x0a) == 0)
				{
					break;
				}
			}

			mStart++;

			//
			// check the hashes
			//
			byte[]  hash = new byte[digest.getDigestSize()];

			digest.doFinal(hash, 0);

			int off = block.Length - delta - hash.Length;

			//
			// there must be at least one byte of message string
			//
			if ((off - mStart) <= 0)
			{
				clearBlock(mBuf);
				clearBlock(block);

				return false;
			}

			for (int i = 0; i != hash.Length; i++)
			{
				block[off + i] ^= hash[i];
				if (block[off + i] != 0)
				{
					clearBlock(mBuf);
					clearBlock(block);

					return false;
				}
			}

			//
			// if we contain the whole message as well, check the hash of that also
			//
			if ((block[0] & 0x20) == 0)
			{
				digest.update(block, mStart, off - mStart);
				digest.doFinal(block, off);
				for (int i = 0; i != hash.Length; i++)
				{
					block[off + i] ^= hash[i];
					if (block[off + i] != 0)
					{
						clearBlock(mBuf);
						clearBlock(block);

						return false;
					}
				}
			}

			clearBlock(mBuf);
			clearBlock(block);

			return true;
		}
	}

}